from collections import Counter
import torch
from typing import List, Tuple
from transformers import AutoModelForCausalLM, AutoTokenizer
import itertools
import heapq

from bcipy.helpers.symbols import BACKSPACE_CHAR, SPACE_CHAR
from bcipy.language.main import LanguageModel, ResponseType

from bcipy.helpers.exceptions import InvalidLanguageModelException

from scipy.special import logsumexp
from scipy.special import softmax

from bcipy.config import LM_PATH


class CausalLanguageModel(LanguageModel):
    """Character language model based on a pre-trained causal model, GPT-2 by default."""

    def __init__(self,
                 response_type: ResponseType,
                 symbol_set: List[str],
                 lang_model_name: str = None,
                 lm_path: str = None,
                 lm_device: str = "cpu",
                 lm_left_context: str = "",
                 ):
        """
        Initialize instance variables and load the language model with given path
        Args:
            response_type - SYMBOL only
            symbol_set    - list of symbol strings
            lang_model_name    - name of the Hugging Face casual language model to load
            lm_path     - load fine-tuned model from specified directory
            lm_device        - device to use for making predictions (cpu, mps, or cuda)
            lm_left_context  - text to condition start of sentence on
        """
        super().__init__(response_type=response_type, symbol_set=symbol_set)
        self.model = None
        self.tokenizer = None
        self.vocab_size = 0
        self.valid_vocab = []
        self.vocab = {}
        self.longest_token = 0
        self.index_to_word = {}
        self.index_to_word_lower = {}
        self.symbol_set_lower = None
        self.device = lm_device
        self.left_context = lm_left_context

        # We optionally load the model from a local directory, but if this is not
        # specified, we load a Hugging Face model

        causal_params = self.parameters['causal']
        self.model_name = lang_model_name or causal_params['model_name']['value']

        local_model_path = lm_path or causal_params['model_path']['value']
        self.model_dir = f"{LM_PATH}/{local_model_path}" if local_model_path != "" else self.model_name

        # parameters for search
        self.beam_width = 8
        self.batch_size = 8

        # backoff to the previous space
        self.token_backoff = -1

        self.load()

    def supported_response_types(self) -> List[ResponseType]:
        return [ResponseType.SYMBOL]

    def _build_vocab(self) -> None:
        """
        Build a vocabulary table mapping token index to word strings
        """

        for i in range(self.vocab_size):
            word = self.tokenizer.decode([i])
            word_lower = word.lower()
            self.index_to_word[i] = word
            self.index_to_word_lower[i] = word_lower
            valid = True
            for ch in word_lower:
                # The space char is only valid once we convert spaces to the space char
                if ch == SPACE_CHAR:
                    valid = False
                    break
                if ch == ' ':
                    continue
                elif ch not in self.symbol_set_lower:
                    valid = False
                    break
            if valid:
                self.valid_vocab += i,
                length = len(word)
                if length > self.longest_token:
                    self.longest_token = length
                for j in range(length):
                    key = word_lower[0:j + 1].replace(' ', SPACE_CHAR)
                    if key not in self.vocab:
                        self.vocab[key] = []
                    self.vocab[key] += i,

        # Get the index we use for the start or end pseudo-word
        if self.left_context == "":
            if "gpt2" in self.model_name:
                self.left_context = "<|endoftext|>"
            else:
                self.left_context = "</s>"
        # Get token id(s) for the left context we condition all sentences on
        self.left_context_tokens = self._encode(self.left_context)

    def _encode(self, text: str) -> List[int]:
        tokens = self.tokenizer.encode(text)
        if len(tokens) > 1 and self.model_name.startswith("facebook/opt"):
            tokens = tokens[1:]

        return tokens

    def predict(self, evidence: List[str]) -> List[Tuple]:
        """
        Given an evidence of typed string, predict the probability distribution of
        the next symbol
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbols with probability
        """

        assert self.model is not None, "language model does not exist!"

        converted_context = "".join(evidence)
        converted_context_lower = converted_context.lower()

        context = converted_context.replace(SPACE_CHAR, ' ')
        context_lower = context.lower()

        # Index in the hypothesis string that is the next character after our context
        target_pos = len(context_lower)

        # Remove token_backoff tokens from the end of the context
        # If token_backoff is -1 or goes beyond a word boundary, then
        # search from the last space character in the context
        # If no space, then from the very beginning
        tokens = self._encode(context_lower)

        # Look for the last space in the context, or -1 if no begin_text in context yet
        pos = context_lower.rfind(" ")
        if pos >= 0:
            truncated_context = context_lower[0:pos]
            truncated_tokens = self._encode(truncated_context)
            # Insert the left context tokens at the start of the sequence
            truncated_tokens[0:0] = self.left_context_tokens
        else:
            # Didn't find space so start inference with just the left context tokens
            truncated_tokens = self.left_context_tokens

        if self.token_backoff == -1 or len(tokens) - self.token_backoff < len(truncated_tokens):
            tokens = truncated_tokens
        else:
            tokens = tokens[:-self.token_backoff]

        # Build up the sequence text for the context
        # Start after the left context tokens
        sequence_text = ""
        for i in range(len(self.left_context_tokens), len(tokens)):
            sequence_text = sequence_text + self.index_to_word_lower[tokens[i]]
        valid = [(0.0, tokens, sequence_text)]
        heapq.heapify(valid)

        # Create a hash mapping each valid following character to a list of log probabilities
        char_to_log_probs = {}

        while len(valid) > 0:
            # Only work on the top hypotheses from the last round of extension
            current = list(valid)

            # Add new extended hypotheses to this list
            valid.clear()

            # Keep going until we have extended all hypotheses in the current set
            while len(current) > 0:
                current_batch = 0
                batch_tensors = []
                batch_sequences = []
                batch_likelihoods = []
                batch_seq_text = []
                while len(current) > 0 and current_batch < self.batch_size:
                    # Get the new sequence to work on
                    (current_likelihood, sequence, sequence_text) = current.pop(0)
                    batch_tensors += torch.tensor(sequence),
                    batch_sequences += sequence,
                    batch_likelihoods += current_likelihood,
                    batch_seq_text += sequence_text,
                    current_batch += 1

                tokens_tensor = torch.stack(tuple(batch_tensors)).to(self.device)

                with torch.no_grad():
                    logits = self.model(tokens_tensor).logits
                    log_probs = torch.log_softmax(logits[:, -1, :], dim=1).to("cpu")

                for j in range(current_batch):
                    sequence_text = batch_seq_text[j]
                    vocab = []
                    extra_vocab = []

                    remaining_context = converted_context_lower[len(sequence_text):]
                    if len(remaining_context) == 0:
                        vocab = self.valid_vocab
                    else:
                        if remaining_context in self.vocab:
                            vocab = self.vocab[remaining_context]
                        for i in range(1, len(remaining_context)):
                            tokenization = self._encode(context_lower[len(sequence_text):len(sequence_text) + i])
                            if len(tokenization) == 1:
                                extra_vocab += tokenization[0],

                    # Create a list of token indexes that are a prefix of target text
                    # We go over all the integer IDs in the vocab and extra_vocab lists
                    for i in itertools.chain(vocab, extra_vocab):
                        hypo_str = sequence_text + self.index_to_word_lower[i]
                        hypo_seq = batch_sequences[j].copy()
                        hypo_seq += i,

                        # Add the log prob of this token to the previous running total
                        # For some reason the float cast makes it run faster
                        likelihood = batch_likelihoods[j] + float(log_probs[j][i])
                        # If we have extended to a space following the context, then that hypothesis gets to be done
                        # This takes a lot longer that just requiring extending beyond existing context
                        # Just require hypotheses to extend beyond the existing typed context
                        if len(hypo_str) > len(context):
                            ch = hypo_str[target_pos]

                            # Create an empty list if we haven't seen this character before
                            if ch not in char_to_log_probs:
                                char_to_log_probs[ch] = []
                            char_to_log_probs[ch] += likelihood,

                        else:
                            hypo = (likelihood, hypo_seq, hypo_str)
                            if len(valid) < self.beam_width:
                                heapq.heappush(valid, hypo)
                            else:
                                heapq.heappushpop(valid, hypo)

        # Parallel array to symbol_set for storing the marginals
        char_probs = []
        for ch in self.symbol_set_lower:
            # Convert space to the underscore used in BciPy
            if ch == SPACE_CHAR:
                target_ch = ' '
            else:
                target_ch = ch

            # Handle cases when symbols are never seen
            if target_ch in char_to_log_probs:
                char_probs += logsumexp(char_to_log_probs[target_ch]),
            else:
                char_probs += float("-inf"),
        # Normalize to a distribution that sums to 1
        char_probs = softmax(char_probs)

        next_char_pred = Counter()

        for i, ch in enumerate(self.symbol_set_lower):
            if ch is SPACE_CHAR:
                next_char_pred[ch] = char_probs[i]
            else:
                next_char_pred[ch.upper()] = char_probs[i]

        next_char_pred[BACKSPACE_CHAR] = 0.0

        return list(sorted(next_char_pred.items(), key=lambda item: item[1], reverse=True))

    def update(self) -> None:
        """Update the model state"""
        ...

    def load(self) -> None:
        """
            Load the language model and tokenizer, initialize class variables
        """
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, use_fast=False)
        except BaseException:
            raise InvalidLanguageModelException(f"{self.model_name} is not a valid model identifier on HuggingFace.")
        self.vocab_size = self.tokenizer.vocab_size
        try:
            self.model = AutoModelForCausalLM.from_pretrained(self.model_dir)
        except BaseException:
            raise InvalidLanguageModelException(
                f"{self.model_dir} is not a valid local folder or model identifier on HuggingFace.")

        self.model.eval()

        self.model.to(self.device)

        self.symbol_set_lower = []
        for ch in self.symbol_set:
            if ch is SPACE_CHAR:
                self.symbol_set_lower.append(SPACE_CHAR)
            elif ch is BACKSPACE_CHAR:
                continue
            else:
                self.symbol_set_lower.append(ch.lower())

        self._build_vocab()

    def state_update(self, evidence: List[str]) -> List[Tuple]:
        """
            Wrapper method that takes in evidence text, and output probability distribution
            of next character
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbol with probability
        """
        next_char_pred = self.predict(evidence)

        return next_char_pred



from collections import Counter
from typing import Dict, List, Tuple
from math import isclose

from bcipy.language.main import LanguageModel, ResponseType

from bcipy.helpers.exceptions import InvalidLanguageModelException

# pylint: disable=unused-import
# flake8: noqa
"""All supported models must be imported"""
from bcipy.language.model.causal import CausalLanguageModel
from bcipy.language.model.kenlm import KenLMLanguageModel
from bcipy.language.model.unigram import UnigramLanguageModel


class MixtureLanguageModel(LanguageModel):
    """
        Character language model that mixes any combination of other models
    """

    supported_lm_types = ["CAUSAL", "UNIGRAM", "KENLM"]

    @staticmethod
    def language_models_by_name() -> Dict[str, LanguageModel]:
        """Returns available language models indexed by name."""
        return {lm.name(): lm for lm in LanguageModel.__subclasses__()}

    @staticmethod
    def validate_parameters(types: List[str], weights: List[float], params: List[Dict[str, str]]):
        if params is not None:
            if (types is None) or (len(types) != len(params)):
                raise InvalidLanguageModelException("Length of parameters does not match length of types")

        if weights is not None:
            if (types is None) or (len(types) != len(weights)):
                raise InvalidLanguageModelException("Length of weights does not match length of types")
            if not isclose(sum(weights), 1.0, abs_tol=1e-05):
                raise InvalidLanguageModelException("Weights do not sum to 1")

        if types is not None:
            if weights is None:
                raise InvalidLanguageModelException("Model weights not provided")
            if params is None:
                raise InvalidLanguageModelException("Model parameters not provided")
            if not all(x in MixtureLanguageModel.supported_lm_types for x in types):
                raise InvalidLanguageModelException(f"Supported model types: {MixtureLanguageModel.supported_lm_types}")

    def __init__(self,
                 response_type: ResponseType,
                 symbol_set: List[str],
                 lm_types: List[str] = None,
                 lm_weights: List[float] = None,
                 lm_params: List[Dict[str, str]] = None):
        """
        Initialize instance variables and load the language model with given path
        Args:
            response_type - SYMBOL only
            symbol_set - list of symbol strings
            lm_types - list of types of models to mix
            lm_weights - list of weights to use when mixing the models
            lm_params - list of dictionaries to pass as parameters for each model's instantiation
        """

        MixtureLanguageModel.validate_parameters(lm_types, lm_weights, lm_params)

        super().__init__(response_type=response_type, symbol_set=symbol_set)
        self.models = list()
        self.response_type = response_type
        self.symbol_set = symbol_set

        mixture_params = self.parameters['mixture']
        self.lm_types = lm_types or mixture_params['model_types']['value']
        self.lm_weights = lm_weights or mixture_params['model_weights']['value']
        self.lm_params = lm_params or mixture_params['model_params']['value']

        MixtureLanguageModel.validate_parameters(self.lm_types, self.lm_weights, self.lm_params)

        self.load()

    def supported_response_types(self) -> List[ResponseType]:
        return [ResponseType.SYMBOL]

    @staticmethod
    def interpolate_language_models(lms: List[Dict[str, float]], coeffs: List[float]) -> List[Tuple]:
        """
        interpolate two or more language models
        Args:
            lms - output from the language models (a list of dicts with char as keys and prob as values)
            coeffs - list of rescale coefficients, lms[0] will be scaled by coeffs[0] and so on
        Response:
            a list of (char, prob) tuples representing an interpolated language model
        """
        combined_lm = Counter()

        for i, lm in enumerate(lms):
            for char in lm:
                combined_lm[char] += lm[char] * coeffs[i]

        return list(sorted(combined_lm.items(), key=lambda item: item[1], reverse=True))

    def predict(self, evidence: List[str]) -> List[Tuple]:
        """
        Given an evidence of typed string, predict the probability distribution of
        the next symbol
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbols with probability
        """

        pred_list = list()

        # Generate predictions from each component language model
        pred_list = [dict(model.predict(evidence)) for model in self.models]

        # Mix the component models
        next_char_pred = MixtureLanguageModel.interpolate_language_models(pred_list, self.lm_weights)

        return next_char_pred

    def update(self) -> None:
        """Update the model state"""
        ...

    def load(self) -> None:
        """
            Load the language models to be mixed
        """

        language_models = MixtureLanguageModel.language_models_by_name()
        for lm_type, params in zip(self.lm_types, self.lm_params):
            model = language_models[lm_type]
            lm = None
            try:
                lm = model(self.response_type, self.symbol_set, **params)
            except InvalidLanguageModelException as e:
                raise InvalidLanguageModelException(f"Error in creation of model type {lm_type}: {e.message}")

            self.models.append(lm)

    def state_update(self, evidence: List[str]) -> List[Tuple]:
        """
            Wrapper method that takes in evidence text, and output probability distribution
            of next character
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbol with probability
        """
        next_char_pred = self.predict(evidence)

        return next_char_pred



"""Defines common functionality for GUI layouts."""
from typing import Tuple
from psychopy import visual


class Container:
    """Protocol for an enclosing container with units and size."""
    win: visual.Window
    size: Tuple[float, float]
    units: str


class WindowContainer(Container):
    """Wrapper for a Window. This is not needed in Python 3.8+ if Container is
    a subclass of typing.Protocol."""

    def __init__(self, win: visual.Window):
        self.win = win

    @property
    def size(self):
        """Window size"""
        return self.win.size

    @property
    def units(self):
        """Window units"""
        return self.win.units


def at_top(parent: Container, height: float):
    """Constructs a layout of a given height that spans the full width of the
    window and is positioned at the top.

    Parameters
    ----------
        height - value in 'norm' units
    """
    return Layout(parent=parent,
                  left=-1.0,
                  top=1.0,
                  right=1.0,
                  bottom=1.0 - height)


class Layout(Container):
    """Base class for a component with methods for positioning elements.
    """

    def __init__(self, parent: Container, left: float, top: float,
                 right: float, bottom: float):

        self.parent = parent
        self.top = top
        self.left = left
        self.bottom = bottom
        self.right = right
        self.check_invariants()

    def check_invariants(self):
        """Check that all invariants hold true."""
        # TODO: units could be configurable; min and max depends on units.
        # https://psychopy.org/general/units.html#units
        assert self.parent.units == "norm", "Position calculations assume norm units."

        assert (0.0 <= self.height <= 2.0), "Height must be in norm units."
        assert (0.0 <= self.width <= 2.0), "Width must be in norm units."
        assert (-1.0 <= self.top <= 1.0), "Top must be a y-value in norm units"
        assert (-1.0 <= self.left <=
                1.0), "Left must be an x-value in norm units"
        assert (-1.0 <= self.bottom <=
                1.0), "Bottom must be a y-value in norm units"
        assert (-1.0 <= self.right <=
                1.0), "Right must be an x-value in norm units"

        assert 0 < self.width <= self.parent.size[
            0], "Width must be greater than 0 and fit within the parent width."
        assert 0 < self.height <= self.parent.size[
            1], "Height must be greater than 0 and fit within the parent height."

    @property
    def win(self) -> visual.Window:
        """Returns the Window on which the layout will be drawn."""
        return self.parent.win

    @property
    def units(self) -> str:
        """Units for drawing; affects calculations related to position, size and radius."""
        return self.parent.units

    @property
    def size(self) -> Tuple[float, float]:
        """Layout size."""
        return (self.width, self.height)

    @property
    def width(self) -> float:
        """Width in norm units of this component."""
        return self.right - self.left

    @property
    def height(self) -> float:
        """Height in norm units of this component."""
        return self.top - self.bottom

    @property
    def horizontal_middle(self) -> float:
        """x-axis value in norm units for the midpoint of this component"""
        return (self.left + self.right) / 2

    @property
    def vertical_middle(self) -> float:
        """x-axis value in norm units for the midpoint of this component."""
        return (self.top + self.bottom) / 2

    @property
    def center(self) -> Tuple[float, float]:
        """Center point of the component in norm units. Returns a (x,y) tuple."""
        return (self.horizontal_middle, self.vertical_middle)

    @property
    def left_middle(self) -> Tuple[float, float]:
        """Point centered on the left-most edge."""
        return (self.left, self.vertical_middle)

    @property
    def right_middle(self) -> Tuple[float, float]:
        """Point centered on the right-most edge."""
        return (self.right, self.vertical_middle)



"""Task bar component"""

from typing import Dict, List
from psychopy import visual
from psychopy.visual.basevisual import BaseVisualStim
from psychopy.visual.line import Line
import bcipy.display.components.layout as layout


class TaskBar:
    """Component for displaying task-related information in a task window. The
    component elements are positioned at the top of the window.

    Parameters
    ----------
        win - visual.Window on which to render elements
        colors - Ordered list of colors to apply to task stimuli
        font - Font to apply to all task stimuli
        height - Height of all task text stimuli
        text - Task text to apply to stimuli
    """

    def __init__(self,
                 win: visual.Window,
                 colors: List[str] = None,
                 font: str = 'Courier New',
                 height: float = 0.1,
                 text: str = ''):
        self.colors = colors or ['white']
        self.font = font
        self.height = height
        self.text = text
        self.layout = layout.at_top(layout.WindowContainer(win),
                                    self.compute_height())
        self.stim = self.init_stim()

    def compute_height(self):
        """Computes the component height using the provided config."""
        padding = 0.05  # determined by trial and error
        return self.height + padding

    def init_stim(self) -> Dict[str, BaseVisualStim]:
        """Initialize the stimuli elements."""
        task = self.text_stim()
        return {'task_text': task, 'border': self.border_stim()}

    def draw(self):
        """Draw the task bar elements."""
        for _key, stim in self.stim.items():
            stim.draw()

    def update(self, text: str = ''):
        """Update the task bar to display the given text."""
        self.stim['task_text'].text = text

    def border_stim(self) -> Line:
        """Create the task bar outline"""
        return visual.line.Line(win=self.layout.win,
                                units=self.layout.units,
                                start=(self.layout.left, self.layout.bottom),
                                end=(self.layout.right, self.layout.bottom),
                                lineColor=self.colors[0])

    def text_stim(self, **kwargs) -> visual.TextStim:
        """Constructs a TextStim. Uses the config to set default properties
        which may be overridden by providing keyword args.
        """

        props = {**self.default_text_props(), **kwargs}
        return visual.TextStim(**props)

    def default_text_props(self) -> dict:
        """Default properties for constructing a TextStim."""
        return {
            'win': self.layout.win,
            'text': self.text,
            'pos': self.layout.center,
            'units': self.layout.units,
            'font': self.font,
            'height': self.height,
            'color': self.colors[0]
        }


class CalibrationTaskBar(TaskBar):
    """Task bar for Calibration tasks. Displays the count of inquiries.

    Parameters
    ----------
        win - visual.Window on which to render elements
        inquiry_count - total number of inquiries to display
        current_index - index of the current inquiry
         **config - display config (colors, font, height)
    """

    def __init__(self,
                 win: visual.Window,
                 inquiry_count: int,
                 current_index: int = 1,
                 **config):
        self.inquiry_count = inquiry_count
        self.current_index = current_index
        super().__init__(win, **config)

    def init_stim(self) -> Dict[str, BaseVisualStim]:
        """Initialize the stimuli elements."""

        task = self.text_stim(text=self.displayed_text(),
                              pos=self.layout.left_middle,
                              anchorHoriz='left',
                              alignText='left')

        return {'task_text': task, 'border': self.border_stim()}

    def update(self, text: str = ''):
        """Update the displayed text"""
        self.current_index += 1
        self.stim['task_text'].text = self.displayed_text()

    def displayed_text(self) -> str:
        """Text to display. Computed from the current_index and inquiry_count. Ex. '2/100'"""
        return f" {self.current_index}/{self.inquiry_count}"


class CopyPhraseTaskBar(TaskBar):
    """Task bar for the Copy Phrase Task

    Parameters
    ----------
        win - visual.Window on which to render elements
        task_text - text for the participant to spell
        spelled_text - text that has already been spelled
        **config - display config (colors, font, height)
    """

    def __init__(self,
                 win: visual.Window,
                 task_text: str = '',
                 spelled_text: str = '',
                 **config):
        self.task_text = task_text
        self.spelled_text = spelled_text
        super().__init__(win, **config)

    def compute_height(self):
        """Computes the component height using the provided config."""
        padding = 0.05  # determined by trial and error
        # height is doubled to account for task_text and spelled_text being on
        # separate lines.
        return (self.height * 2) + padding

    def init_stim(self) -> Dict[str, BaseVisualStim]:
        """Initialize the stimuli elements."""

        task = self.text_stim(text=self.task_text,
                              pos=self.layout.center,
                              anchorVert='bottom')

        spelled = self.text_stim(text=self.displayed_text(),
                                 pos=self.layout.center,
                                 color=self.colors[-1],
                                 anchorVert='top')

        return {
            'task_text': task,
            'spelled_text': spelled,
            'border': self.border_stim()
        }

    def update(self, text: str = ''):
        """Update the task bar to display the given text."""
        self.spelled_text = text
        self.stim['spelled_text'].text = self.displayed_text()

    def displayed_text(self):
        """Spelled text padded for alignment."""
        diff = len(self.task_text) - len(self.spelled_text)
        if (diff > 0):
            return self.spelled_text + (' ' * diff)
        return self.spelled_text



from typing import List, Tuple
from bcipy.helpers.symbols import BACKSPACE_CHAR, SPACE_CHAR
from bcipy.language.main import LanguageModel, ResponseType
from bcipy.helpers.exceptions import InvalidLanguageModelException
import json
from bcipy.config import LM_PATH


class UnigramLanguageModel(LanguageModel):
    """Character language model based on trained unigram weights"""

    def __init__(self, response_type: ResponseType, symbol_set: List[str], lm_path: str = None):

        super().__init__(response_type=response_type, symbol_set=symbol_set)
        self.model = None
        self.lm_path = lm_path or f"{LM_PATH}/unigram.json"

        try:
            with open(self.lm_path) as json_file:
                self.unigram_lm = json.load(json_file)
        except BaseException:
            raise InvalidLanguageModelException("Unable to load Unigram model from file")

        self.unigram_lm[SPACE_CHAR] = self.unigram_lm.pop("SPACE_CHAR")
        self.unigram_lm[BACKSPACE_CHAR] = self.unigram_lm.pop("BACKSPACE_CHAR")

        if not set(self.unigram_lm.keys()) == set(self.symbol_set):
            raise InvalidLanguageModelException("Invalid unigram model symbol set!")

        self.unigram_lm = sorted(self.unigram_lm.items(), key=lambda item: item[1], reverse=True)

        self.load()

    def supported_response_types(self) -> List[ResponseType]:
        return [ResponseType.SYMBOL]

    def predict(self, evidence: List[str]) -> List[Tuple]:
        """
        Given an evidence of typed string, predict the probability distribution of
        the next symbol
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbols with probability
        """

        return self.unigram_lm

    def update(self) -> None:
        """Update the model state"""
        ...

    def load(self) -> None:
        """
            Load the language model and tokenizer, initialize class variables
        Args:
            path: language model file path, can be just "unigram"
        """

    def state_update(self, evidence: List[str]) -> List[Tuple]:
        """
            Wrapper method that takes in evidence text, and output probability distribution
            of next character - evidence does not matter for unigram model
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbol with probability
        """
        return self.unigram_lm



from collections import Counter
from typing import List, Tuple
from bcipy.helpers.symbols import BACKSPACE_CHAR, SPACE_CHAR
from bcipy.language.main import LanguageModel, ResponseType
from bcipy.helpers.exceptions import InvalidLanguageModelException, KenLMInstallationException
from bcipy.config import LM_PATH
try:
    import kenlm
except BaseException:
    raise KenLMInstallationException(
        "Please install the requisite kenlm package:\n'pip install kenlm==0.1 --global-option=\"--max_order=12\"")
import numpy as np


class KenLMLanguageModel(LanguageModel):
    """Character n-gram language model using the KenLM library for querying"""

    def __init__(self, response_type: ResponseType, symbol_set: List[str], lm_path: str = None):

        super().__init__(response_type=response_type, symbol_set=symbol_set)
        self.model = None
        kenlm_params = self.parameters['kenlm']
        kenlm_model = kenlm_params['model_file']['value']
        self.lm_path = lm_path or f"{LM_PATH}/{kenlm_model}"

        self.load()

    def supported_response_types(self) -> List[ResponseType]:
        return [ResponseType.SYMBOL]

    def predict(self, evidence: List[str]) -> List[Tuple]:
        """
        Given an evidence of typed string, predict the probability distribution of
        the next symbol
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbols with probability
        """

        # Do not modify the original parameter, could affect mixture model
        context = evidence.copy()

        if len(context) > 11:
            context = context[-11:]

        evidence_str = ''.join(context).lower()

        for i, ch in enumerate(context):
            if ch == SPACE_CHAR:
                context[i] = "<sp>"

        self.model.BeginSentenceWrite(self.state)

        # Update the state one token at a time based on evidence, alternate states
        for i, token in enumerate(context):
            if i % 2 == 0:
                self.model.BaseScore(self.state, token.lower(), self.state2)
            else:
                self.model.BaseScore(self.state2, token.lower(), self.state)

        next_char_pred = None

        # Generate the probability distribution based on the final state
        if len(context) % 2 == 0:
            next_char_pred = self.prob_dist(self.state)
        else:
            next_char_pred = self.prob_dist(self.state2)

        return next_char_pred

    def update(self) -> None:
        """Update the model state"""
        ...

    def load(self) -> None:
        """
            Load the language model, initialize state variables
        Args:
            path: language model file path
        """

        try:
            self.model = kenlm.LanguageModel(self.lm_path)
        except BaseException:
            raise InvalidLanguageModelException(
                f"A valid model path must be provided for the KenLMLanguageModel.\nPath{self.lm_path} is not valid.")

        self.state = kenlm.State()
        self.state2 = kenlm.State()

    def state_update(self, evidence: List[str]) -> List[Tuple]:
        """
            Wrapper method that takes in evidence text and outputs probability distribution
            of next character
        Args:
            evidence - a list of characters (typed by the user)
        Response:
            A list of symbols with probabilities
        """
        next_char_pred = self.predict(evidence)

        return next_char_pred

    def prob_dist(self, state: kenlm.State) -> List[Tuple]:
        """
            Take in a state and generate the probability distribution of next character
        Args:
            state - the kenlm state updated with the evidence
        Response:
            A list of symbols with probability
        """
        next_char_pred = Counter()

        temp_state = kenlm.State()

        for char in self.symbol_set:
            # Backspace probability under the LM is 0
            if char == BACKSPACE_CHAR:
                next

            score = 0.0

            # Replace the space character with KenLM's <sp> token
            if char == SPACE_CHAR:
                score = self.model.BaseScore(state, '<sp>', temp_state)
            else:
                score = self.model.BaseScore(state, char.lower(), temp_state)

            # BaseScore returns log probs, convert by putting 10 to its power
            next_char_pred[char] = pow(10, score)

        sum = np.sum(list(next_char_pred.values()))
        for char in self.symbol_set:
            next_char_pred[char] /= sum

        return list(sorted(next_char_pred.items(), key=lambda item: item[1], reverse=True))



"""Uniform language model"""
from typing import Dict, List, Tuple, Union, Optional

import numpy as np

from bcipy.language.main import LanguageModel, ResponseType


class UniformLanguageModel(LanguageModel):
    """Language model in which probabilities for symbols are uniformly
    distributed.

    Parameters
    ----------
        response_type - SYMBOL only
        symbol_set - optional specify the symbol set, otherwise uses DEFAULT_SYMBOL_SET
    """

    def __init__(self,
                 response_type: Optional[ResponseType] = None,
                 symbol_set: Optional[List[str]] = None):
        super().__init__(response_type=response_type, symbol_set=symbol_set)

    def supported_response_types(self) -> List[ResponseType]:
        return [ResponseType.SYMBOL]

    def predict(self, evidence: Union[str, List[str]]) -> List[Tuple]:
        """
        Using the provided data, compute probabilities over the entire symbol.
        set.

        Parameters
        ----------
            evidence  - list of previously typed symbols

        Returns
        -------
            list of (symbol, probability) tuples
        """
        probs = equally_probable(self.symbol_set)
        return list(zip(self.symbol_set, probs))

    def update(self) -> None:
        """Update the model state"""

    def load(self) -> None:
        """Restore model state from the provided checkpoint"""


def equally_probable(alphabet: List[str],
                     specified: Dict[str, float] = None) -> List[float]:
    """Returns a list of probabilities which correspond to the provided
    alphabet. Unless overridden by the specified values, all items will
    have the same probability. All probabilities sum to 1.0.

    Parameters:
    ----------
        alphabet - list of symbols; a probability will be generated for each.
        specified - dict of symbol => probability values for which we want to
            override the default probability.
    Returns:
    --------
        list of probabilities (floats)
    """
    n_letters = len(alphabet)
    if not specified:
        return np.full(n_letters, 1 / n_letters)

    # copy specified dict ignoring non-alphabet items
    overrides = {k: specified[k] for k in alphabet if k in specified}
    assert sum(overrides.values()) < 1

    prob = (1 - sum(overrides.values())) / (n_letters - len(overrides))
    # override specified values
    return [overrides[sym] if sym in overrides else prob for sym in alphabet]
